import re
import importlib.machinery
import importlib.util
import logging

from redpesk_utils.log import red_log

class NoBumpReleaseError(Exception):
    pass

# Import bumpspec from python script
try:
    bumspec_path="/usr/bin/rpmdev-bumpspec"
    loader = importlib.machinery.SourceFileLoader('bumpspec', bumspec_path)
    spec = importlib.util.spec_from_loader('bumpspec', loader)
    bumpspec = importlib.util.module_from_spec(spec)
    try:
        spec.loader.exec_module(bumpspec)
    except SystemExit:
        pass
except Exception:
    raise NoBumpReleaseError(f"Cannot import python script {bumspec_path} as module"
            ": Need to install rpmdevtools""")

class SpecFileBumpSetRelease(bumpspec.SpecFile):
    _re_main = r'^(0.)?\d+'
    _re_fallback = r'\d+$'

    def __init__(self, filename, latest_release, logger):
        super().__init__(filename)
        self.latest_release = latest_release
        self.fallback = False
        self.logger = logger

    def _incr_decr(self, val, incr):
        if self.fallback:
            return val + 1

        # use same behavior as rpmdev-bumpspec for release that start with '0.'
        # it increments the 'minor' release number
        return val + (1 if incr else -1)*(1 if val >= 1 else 0.1)

    def _incr(self, val):
        return self._incr_decr(val, True)

    def _decr(self, val):
        return self._incr_decr(val, False)

    def _main_release(self, release):
        new = super().increaseMain(release)
        relval = float(re.search(SpecFileBumpSetRelease._re_main, new).group())
        if relval.is_integer():
            relval = int(relval)
        return new, self._decr(relval)

    def _fallback_release(self, release):
        new = super().increaseFallback(release)
        relval = int(re.search(SpecFileBumpSetRelease._re_fallback, new).group()) - 1
        return new, relval

    def _increasewithlatest(self, regex, new, relval, relvallatest):
        if relval > relvallatest:
            raise NoBumpReleaseError(f"Do not bump specfile: relval={relval} is greater than relvallatest={relvallatest}")

        relnew = re.sub(regex, str(self._incr(relvallatest)), new)
        self.logger.info("new release is %s", relnew)
        return relnew

    def increaseMain(self, release):
        self.logger.info("release found is %s", release)
        new, relval = self._main_release(release)
        self.logger.debug("new=%s relval=%s", new, relval)
        _, relvallatest = self._main_release(self.latest_release)
        relnew = new

        return self._increasewithlatest(SpecFileBumpSetRelease._re_main, new, relval, relvallatest)

    def increaseFallback(self, release):
        # in fallback mode call mother increaseMain
        self.increaseMain = super().increaseMain
        self.fallback = True
        new, relval = self._fallback_release(release)

        relvallatest = None
        _, relvallatest = self._fallback_release(self.latest_release)

        return self._increasewithlatest(SpecFileBumpSetRelease._re_fallback, new, relval, relvallatest)

def modify_release(specfile, new_release, logger=red_log):
    spec = SpecFileBumpSetRelease(specfile, new_release, logger=red_log)
    try:
        spec.bumpRelease()
    except NoBumpReleaseError:
        logger.info("No need to bump release")
        return
    spec.writeFile(specfile)
