# -*- coding: utf-8 -*-
# vim:expandtab:autoindent:tabstop=4:shiftwidth=4:filetype=python:textwidth=0:
# cross_init - a Python library for Redpesk SDK
#
# Copyright (C) 2020 IoT.bzh Company
# Contact: https://www.iot.bzh/licensing
#
# Author(s): Romain Forlot <romain.forlot@iot.bzh>
#            Clément Bénier <clement.benierc@iot.bzh>
#
# This file is part of the RedPesk project.
#
# $RP_BEGIN_LICENSE$
# Commercial License Usage
#  Licensees holding valid commercial IoT.bzh licenses may use this file in
#  accordance with the commercial license agreement provided with the
#  Software or, alternatively, in accordance with the terms contained in
#  a written agreement between you and The IoT.bzh Company. For licensing terms
#  and conditions see https://www.iot.bzh/terms-conditions. For further
#  information use the contact form at https://www.iot.bzh/contact.
#
# GNU General Public License Usage
#  Alternatively, this file may be used under the terms of the GNU General
#  Public license version 3. This license is as published by the Free Software
#  Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
#  of this file. Please review the following information to ensure the GNU
#  General Public License requirements will be met
#  https://www.gnu.org/licenses/gpl-3.0.html.
# $RP_END_LICENSE$
import io
import os
import json
import typing
import logging

#########
# CONST #
#########
ENCODE = "utf-8"
GIT_DIR = ".git"
GIT_HEAD = "HEAD"
VERSION = "1.6.0"
SPEC_EXT = ".spec"
JSON_TYPE = "json"
SERVICE = "service"
BUILTINS = "builtins"
METADATA = "metadata"
CHUNK_SIZE = 64 * 1024
DEFAULT_BRANCH = "main"
METADATA_VERSION = "1.1.0"
JSON_EXT = f".{JSON_TYPE}"
REDPESK_FOLDER = ".redpesk"
AUTOGEN_PREFIX = "autogen::"
GIT_COMMIT_UPDATE = "update"
NO_CACHE = "private, max-age=0, no-cache"
SERVICE_FILE = AUTOGEN_PREFIX + SERVICE + JSON_EXT
METADATA_FILE = AUTOGEN_PREFIX + METADATA + JSON_EXT
SERVICE_PATH = os.path.join(REDPESK_FOLDER, SERVICE_FILE)
METADATA_PATH = os.path.join(REDPESK_FOLDER, METADATA_FILE)
TEXT_CHARS = bytearray(
    {7, 8, 9, 10, 12, 13, 27} | set(range(0x20, 0x100)) - {0x7f})


def BYTES_BINARY(
    bytes: bytes) -> bool: return bool(bytes.translate(None, TEXT_CHARS))

def STREAM_BINARY(stream: io.IOBase, content: bool) -> bool:
    chunk = stream.read(1024)
    stream.seek(0)
    return (
        isinstance(chunk, bytes)
        if not content
        else not isinstance(chunk, str) and BYTES_BINARY(chunk)
    )



#########
# FIELD #
#########
FIELD_NAME = "name"
FIELD_TYPE = "type"
FIELD_DATA = "data"
FIELD_VALUE = "value"
FIELD_VALUES = "values"
FIELD_VERSION = "version"
FIELD_REVISION = "revision"
FIELD_PARAMETERS = "parameters"
FIELD_SOURCES_URI = "sources_uri"
FIELD_DESCRIPTION = "description"
FIELD_SPECFILE_LOCATION = "specfile_location"

#########
# REGEX #
#########
REGEX_SHA = r"^[0-9a-f]{40}$"
REGEX_NONGROUPTEXT = r"^[\w\-\/]*$"
REGEX_NONSPECIALTEXT = r"^[\w\-\+\.]*$"
REGEX_UUID = r"[a-f\d]{8}-[a-f\d]{4}-[a-f\d]{4}-[a-f\d]{4}-[a-f\d]{12}"
REGEX_ARCHIVE_VERSION = r"(\d+\.\d+.\d+(?:\+\d{8}(?:\+\d{6})?\+\d+\+g\w{0,64})*)"

#########
# TYPES #
#########
TYPE_INT = "int"
TYPE_FLOAT = "float"
TYPE_BOOL = "boolean"
TYPE_STRING = "string"
TYPE_BUILD_TYPE = "build_type"
TYPE_STRING_LIST = "string_list"
TYPE_STRING_ARRAY = "string_array"
TYPE_SPECFILE_TYPE = "specfile_type"

##############
# BUILD_TYPE #
##############
BUILD_TYPE_ACCEL_SHORT = "accel"
BUILD_TYPE_STANDARD_SHORT = "cross"
BUILD_TYPE_EMULATED_SHORT = "emulated"
BUILD_TYPE_EMULATED = "emulated_build"
BUILD_TYPE_ACCEL = "accel_build_cross"
BUILD_TYPE_STANDARD = "standard_build_cross"
BUILD_TYPE_DEFAULT = BUILD_TYPE_STANDARD
BUILD_TYPE_TRANSLATION = {
    BUILD_TYPE_ACCEL: BUILD_TYPE_ACCEL_SHORT,
    BUILD_TYPE_STANDARD: BUILD_TYPE_STANDARD_SHORT,
    BUILD_TYPE_EMULATED: BUILD_TYPE_EMULATED_SHORT
}
def BUILD_TYPE_SHORT(build_type: str) -> str:
    return next((value for key, value in BUILD_TYPE_TRANSLATION.items() if key == build_type), build_type)

def BUILD_TYPE_LONG(build_type: str) -> str:
    return next((key for key, value in BUILD_TYPE_TRANSLATION.items() if value == build_type), build_type)

#################
# SPECFILE_TYPE #
#################
SPECFILE_TYPE_INTERNAL = "internal"
SPECFILE_TYPE_INTERNAL_SHORT = "internal"
SPECFILE_TYPE_EXTERNAL_URL = "external_url"
SPECFILE_TYPE_EXTERNAL_URL_SHORT = "external"
SPECFILE_TYPE_RELATIVE_PATH = "relative_path"
SPECFILE_TYPE_RELATIVE_PATH_SHORT = "relative"
SPECFILE_TYPE_DEFAULT = SPECFILE_TYPE_INTERNAL
SPECFILE_TYPE_TRANSLATION = {
    SPECFILE_TYPE_INTERNAL: SPECFILE_TYPE_INTERNAL_SHORT,
    SPECFILE_TYPE_EXTERNAL_URL: SPECFILE_TYPE_EXTERNAL_URL_SHORT,
    SPECFILE_TYPE_RELATIVE_PATH: SPECFILE_TYPE_RELATIVE_PATH_SHORT
}
def SPECFILE_TYPE_SHORT(specfile_type: str) -> str:
    return next((value for key, value in SPECFILE_TYPE_TRANSLATION.items() if key == specfile_type), specfile_type)

def SPECFILE_TYPE_LONG(specfile_type: str) -> str:
    return next((key for key, value in SPECFILE_TYPE_TRANSLATION.items() if value == specfile_type), specfile_type)

############
# METADATA #
############
METADATA_BUILD_TYPE_DESCRIPTION = "The way in which the package is built"
METADATA_SPECFILE_TYPE_DESCRIPTION = "The way in which the specfile is handled"
METADATA_SOURCE_URI_DESCRIPTION = "Git URI the autogenerated archive is based on"
METADATA_REVISION_DESCRIPTION = "Git reference the autogenerated archive is based on"
METADATA_SPECFILE_LOCATION_DESCRIPTION = "The path in which the specfile can be found"

##########
# MACROS #
##########
MACRO_NAME = "Name"
MACRO_HEXSHA = "#Hexsha"
MACRO_VERSION = "Version"

##########
# CENTOS #
##########
CENTOS_SPECS_DIR = "SPECS"
CENTOS_SOURCES_DIR = "SOURCES"

###########
# COLOURS #
###########
COLOUR_RESET = "\x1b[0m"
COLOUR_RED = "\x1b[1;31m"
COLOUR_BLUE = "\x1b[1;34m"
COLOUR_WHITE = "\x1b[0;37m"
COLOUR_GREEN = "\x1b[1;32m"
COLOUR_YELLOW = "\x1b[1;33m"

##################
# LOGGING COLOUR #
##################
LOGGING_COLOUR = {
    logging.ERROR: COLOUR_RED,
    logging.INFO: COLOUR_GREEN,
    logging.DEBUG: COLOUR_WHITE,
    logging.WARNING: COLOUR_YELLOW
}

###########
# LOGGING #
###########
LOGGING_SEPARATOR = " | "
LOGGING_METADATA = "METADATA"
LOGGING_FORMAT = '[%(levelname)s][%(name)s][%(module)s:%(funcName)s:%(lineno)d] %(message)s'


def IS_BUILTINS(
    obj: typing.Any) -> bool: return obj.__class__.__module__ == BUILTINS and not isinstance(
        obj,
    typing.Callable)


def encode_dict(dict: typing.Dict[str, typing.Any]) -> str:
    return json.dumps(dict, indent=4, sort_keys=True)

class Project:
    def __init__(self) -> None:
        return