# -*- coding: utf-8 -*-
# vim:expandtab:autoindent:tabstop=4:shiftwidth=4:filetype=python:textwidth=0:
# cross_init - a Python library for Redpesk SDK
#
# Copyright (C) 2020 IoT.bzh Company
# Contact: https://www.iot.bzh/licensing
#
# Author(s): Romain Forlot <romain.forlot@iot.bzh>
#            Clément Bénier <clement.benierc@iot.bzh>
#
# This file is part of the RedPesk project.
#
# $RP_BEGIN_LICENSE$
# Commercial License Usage
#  Licensees holding valid commercial IoT.bzh licenses may use this file in
#  accordance with the commercial license agreement provided with the
#  Software or, alternatively, in accordance with the terms contained in
#  a written agreement between you and The IoT.bzh Company. For licensing terms
#  and conditions see https://www.iot.bzh/terms-conditions. For further
#  information use the contact form at https://www.iot.bzh/contact.
#
# GNU General Public License Usage
#  Alternatively, this file may be used under the terms of the GNU General
#  Public license version 3. This license is as published by the Free Software
#  Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
#  of this file. Please review the following information to ensure the GNU
#  General Public License requirements will be met
#  https://www.gnu.org/licenses/gpl-3.0.html.
# $RP_END_LICENSE$

import os
import re
import grp
import pwd
import stat
import glob
import magic
import shutil
import logging
import fileinput
import subprocess

from pathlib import Path
from os import listdir
from os.path import islink, isdir, join, abspath
from .crossutils import get_arch_from_rpmmacro

logger = logging.getLogger('Crossbuild')
logger.setLevel(logging.DEBUG)


def getLog():
    return logger


def symstartswithsep(path):
    return os.readlink(path).startswith(os.sep)


def list_linkfiles_recursively(dir_path):
    list_linkfiles = []
    getLog().debug(listdir(dir_path))
    for f in listdir(dir_path):
        path_f = join(dir_path, f)
        if isdir(path_f):
            getLog().debug("isdir")
            list_linkfiles += list_linkfiles_recursively(path_f)
        else:
            if islink(path_f) and symstartswithsep(path_f):
                list_linkfiles.append(path_f)
    getLog().debug("list_linkfiles... {}".format(list_linkfiles))
    return list_linkfiles


def resimlinkabsolutepath(sysroot):
    dirs_lib = ["etc/alternatives", "usr/lib", "usr/lib64", "usr/lib32",
                "usr/bin", "usr/libexec"]
    for d in dirs_lib:
        lib = join(sysroot, d)
        if not isdir(lib):
            continue

        file_links = list_linkfiles_recursively(lib)
        getLog().debug("file_links: {}".format(file_links))
        for file_link in file_links:
            simlink = os.readlink(file_link)
            getLog().debug("simlink: {}".format(simlink))
            prefix = ""
            nb_back_dir = len(file_link.split(os.sep)) - \
                len(sysroot.split(os.sep)) - 1
            getLog().debug(nb_back_dir)
            for nb_dir in range(0, nb_back_dir):
                prefix = join(prefix, "..")
            new_link = prefix + simlink
            getLog().debug("new_link: {}".format(new_link))
            getLog().debug("ln -s {} {}".format(new_link, file_link))
            os.remove(file_link)
            os.symlink(new_link, file_link)


def copyandbackupifexists(src, dest):
    if os.path.isfile(src) and os.path.isfile(dest):
        shutil.copy2(dest, '%s_x86_64_backup' % (dest))
        shutil.copy2(src, dest)
    else:
        getLog().debug('File %s or %s does not exists. Pass.' % (src, dest))


class CrossBuildEnv(object):
    # pylint: disable=too-few-public-methods
    """Initialize your environment to be able to crossbuild a certain archs"""

    def __init__(self, target_arch, target_sysroot, rootdir='/', config=None):
        if config:
            self.config = config
        else:
            self.config = {'rpmbuild_arch': target_arch,
                           'target_arch': target_arch}
        self.magic = magic.open(magic.NONE)
        self.magic.load()
        self.sysroot = target_sysroot
        self.chroot = rootdir
        if not isdir(self.chroot):
            logger.info("no chroot in {}".format(self.chroot))
            return
        if not isdir(self.sysroot):
            logger.info("no sysroot in {}".format(self.sysroot))
            return
        self.sysroot_chroot = os.path.join("/", self.sysroot.replace(self.chroot, ""))

        logger.info("Crossbuild plugin loaded with rpmbuild_arch at init: {}".format(
            self.config['rpmbuild_arch']))
        logger.info("Crossbuild plugin loaded with target_arch at init: {}".format(
            self.config['target_arch']))
        logger.info("Crossbuild chroot: {}".format(self.chroot))
        logger.info("Crossbuild target sysroot: {}".format(self.sysroot))
        logger.info("Crossbuild sysroot in chroot: {}".format(self.sysroot_chroot))
        logger.info("Crossbuild target arch: {}".format(
            self.config['rpmbuild_arch']))

        # 'copytree' fails if folder already exist.
        # It can be the case if mockcrossbuild is run twice or more.
        if not os.path.isdir(self.sysroot+'/etc'):
            shutil.copytree(self.chroot+'/etc', self.sysroot +
                            '/etc', symlinks=True)
        else:
            logger.info("{}/etc already exist".format(self.sysroot))

        self.rpm_arch = get_arch_from_rpmmacro(self.config['rpmbuild_arch'])

        resimlinkabsolutepath(self.sysroot)
        logger.info("Crossbuild after resymlink")
        self._fix_config_bin()
        logger.info("Crossbuild after config_bin")
        self._fix_fortran_env()
        logger.info("Crossbuild after fortran")
        self._fix_python_env()
        logger.info("Crossbuild after python")
        self._fix_perl_env()
        logger.info("Crossbuild after perl")
        self._fix_arm()
        self._symlink_ld_linux()

    def get_libdir(self, root):
        macros_path = "{}/usr/lib/rpm/macros:{}/usr/lib/rpm/platform/{}-linux/macros:{}/etc/rpm/macros.{}:{}/builddir/.rpmmacros".format(
            root, root, self.rpm_arch, root, self.rpm_arch, root)
        cmd = ["/usr/bin/rpm", "--macros", macros_path, "-E", "%_libdir"]
        output = subprocess.check_output(cmd, cwd=root)
        return output.decode('utf-8').strip()

    def _fix_arm(self):
        if self.rpm_arch != "arm":
            return

        gnu_soft_path = os.path.join(
            self.sysroot, "usr/include/gnu/stubs-soft.h")
        if not os.path.exists(gnu_soft_path):
            os.mknod(gnu_soft_path, mode=0o644)

    def _replace_shebang(self, afile, encoding):
        internal_sysroot = '{}'.format(self.sysroot.replace(self.chroot, ''))
        content = ''
        encoding = re.sub(r'^(iso|ISO)-', 'iso', encoding)
        with open(afile, 'r+', encoding=encoding) as pyfile:
            content = pyfile.read()
            replaced = re.sub(
                r'^#!\s?(/usr/(?:bin|libexec)/)([\w-]*)', '#!{}\\1\\2'.format(internal_sysroot), content)
            pyfile.seek(0)
            pyfile.truncate()
            pyfile.write(replaced)
        with open('{}.bak'.format(afile), 'w', encoding=encoding) as pyfile:
            pyfile.write(content)

    def _fix_shebang(self, ftype, path):
        if not isdir(path):
            logger.error("fix_shebang: {} is not a dir".format(path))
            return

        with os.scandir(path) as it:
            for entry in it:
                if not entry.name.startswith('.') and entry.is_file() \
                        and self.magic.file(entry.path).startswith(ftype):
                    self._replace_shebang(
                        entry.path, self.magic.file(entry.path).split()[2])

    def _fix_platform_python_config_file(self, config):
        if config.exists():
            pyconfig_resolved = config.resolve().__str__()
            with fileinput.FileInput(pyconfig_resolved, inplace=True, backup='.bak') as pyconfig:
                for line in pyconfig:
                    line = re.sub(r"exec /usr/libexec/platform-python",
                                  r"exec ${CROSS_ROOT}/usr/libexec/platform-python", line)
                    print(line.strip(os.linesep))

    def _fix_python_env(self):
        libdir = self.get_libdir(self.chroot)
        sysconfigdata_regex = r'_sysconfigdata_.*.py$'
        for root, _, f in os.walk('{}{}'.format(self.sysroot, libdir)):
            for f2c in [elt for elt in f if re.search(sysconfigdata_regex, elt)]:
                with fileinput.FileInput(os.path.join(root, f2c), inplace=True, backup='.bak') as f_sysconfig:
                    for line in f_sysconfig:
                        line = line.strip(os.linesep)
                        line = re.sub(
                            r"'g\+\+\s", "'{}-linux-gnu-g++ ".format(self.rpm_arch), line)
                        line = re.sub(
                            r"'gcc\s", "'{}-linux-gnu-gcc ".format(self.rpm_arch), line)
                        line = re.sub(r"(\W*'INCLUDEPY'\W*:\W*')(/usr/include)", r"\1{}\2".format(
                            self.sysroot_chroot), line)
                        print(line)

        self._fix_shebang('Python script', '{}/usr/bin'.format(self.sysroot))

        # Fix the config binary, the automatic wrapper generator function _fix_config_bin do not fit well because
        # it calls a binary that do not exists in buildroot directly either you call it from sysroot or not
        config = Path(
            '{}/usr/libexec/platform-python-config'.format(self.chroot)) # Fix native
        self._fix_platform_python_config_file(config)
        
        config = Path(
            '{}/usr/libexec/platform-python-config'.format(self.sysroot)) # Fix cross
        self._fix_platform_python_config_file(config)

    def _fix_perl_env(self):
        libdir = self.get_libdir(self.chroot)

        copyandbackupifexists('{}{}/perl5/Config.pm'.format(self.sysroot, libdir),
                              '{}{}/perl5/Config.pm'.format(self.chroot, libdir))
        copyandbackupifexists('{}{}/perl5/Config_heavy.pl'.format(self.sysroot,
                                                                  libdir), '{}{}/perl5/Config_heavy.pl'.format(self.chroot, libdir))
        copyandbackupifexists('{}{}/perl5/Errno.pm'.format(self.sysroot, libdir),
                              '{}{}/perl5/Errno.pm'.format(self.chroot, libdir))

    def _fix_fortran_env(self):
        gfortran_file = self.chroot + \
            '/usr/bin/{}-redhat-linux-gnu-gfortran'.format(self.rpm_arch)
        gfortran = '{}/usr/bin/{}-linux-gnu-gfortran'.format(
            self.chroot, self.rpm_arch)
        if os.path.isfile(gfortran) and not os.path.isfile(gfortran_file):
            os.link(gfortran, gfortran_file)

    def _fix_config_bin(self):
        cmd = ["/usr/bin/find", self.chroot+"/usr/bin", "-mmin", "+60", "-name", "*_config", "-o", "-name", "*-config", "!", "-name",
               "x86_64-*-config", "-a", "!", "-name", "cross-*-config", "-a", "!", "-name", "pkg-config", "-a", "!", "-type", "l"]
        output = subprocess.check_output(cmd, cwd=self.chroot)
        config_bin = list(filter(None, output.decode('utf-8').split('\n')))

        with fileinput.FileInput(self.chroot+'/usr/bin/ldd', inplace=True, backup='.bak') as ldd:
            for line in ldd:
                print(re.sub(r'^(RTLDLIST=.*)"$', r'\1 /{}"'.format(
                             self.chroot), line.strip('\n')))

        logger.debug('Crossbuild: Will override {}'.format(config_bin))
        for b in config_bin:
            baseb = os.path.basename(b)
            shutil.move(
                b, "{}/x86_64-linux-gnu-{}".format(os.path.dirname(b), baseb))
            with open(b, 'w') as fd:
                fd.write('#!/bin/bash\n')
                fd.write(
                    'exec ${CROSS_ROOT}${0} $@ | sed -r -e \'s:(-I|^)/usr/include:\\1\'${CROSS_ROOT}\'/usr/include:g\' -e \'s:(-L|^)/usr/lib:\\1\'${CROSS_ROOT}\'/usr/lib:g\' -e \'s:^(/usr)$:\'$CROSS_ROOT\'\1:\'\n')
            os.chmod(b, stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)

        unameb = '{}/usr/bin/uname'.format(self.chroot)
        baseb = os.path.basename(unameb)
        new_uname = "{}/x86_64-linux-gnu-{}".format(
            os.path.dirname(unameb), baseb)
        if not os.path.exists(new_uname):
            logger.info("Create uname wrapper for {}.".format(
                self.config['rpmbuild_arch']))
            shutil.move(unameb, new_uname)
            with open(unameb, 'w') as fd:
                fd.write('#!/bin/bash\n')
                fd.write('if [ "${CROSS_ARCH}" ]; then\n')
                fd.write('exec x86_64-linux-gnu-'+baseb +
                         ' $@ | sed -r \'s:x86_64:\'${CROSS_ARCH}\':g\'\n')
                fd.write('else\n')
                fd.write('exec x86_64-linux-gnu-'+baseb+' $@\n')
                fd.write('fi\n')
            os.chmod(unameb, stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)
        else:
            logger.info("{} already exist".format(new_uname))

    def _symlink_ld_linux(self):
        sysroot_lib = join(self.sysroot, "lib")
        if not isdir(sysroot_lib):
            logger.error(
                "_symlink_ld_linux: {} is not a dir".format(sysroot_lib))
            return
        ldlinux_target = None
        for f in listdir(sysroot_lib):
            if f.startswith("ld-linux-"):
                ldlinux_target = join(sysroot_lib, f)
                if not self.chroot.__eq__('/'):
                    ldlinux_target = ldlinux_target.replace(self.chroot, "")
                break
        if not ldlinux_target:
            return
        new_link = join(self.chroot, "lib", os.path.basename(ldlinux_target))
        if not os.path.isfile(ldlinux_target) and os.path.realpath(new_link) != ldlinux_target:
            os.symlink(ldlinux_target, new_link)
