import types
import sys
import os
import koji
import json
import rpm
import time
import subprocess
import re

from . import load_source

from .crossutils import (NATIVE_ARCH, getCrossArch, isCross, isNative, getArchCrossNative,
                         getArchCrossTarget, getPrjDist, get_arch_from_rpmmacro, SIMPLE,
                         CROSS_SIMPLE, CROSS_COMPILERS, ACCEL)

class MockConfigRedError(Exception):
    pass

class Host:
    def newBuildRoot(self, *arg, **kwargs):
        return 1


def writeMockConfigIgnore(self, *arg, **kwargs):
    pass

def _check_signing(buildroot):
    cmd_prefix=""
    mockuser = getattr(buildroot.options, 'mockuser', None)
    if mockuser:
        cmd_prefix="su -c {}".format(mockuser)

    for macro in ("%{?_signature}", "%{?_gpg_path}", "%{?_gpg_name}", "%{?_gpgbin}", "%{?__gpg_sign_cmd}"):
        if not subprocess.getoutput("{} 'rpm -E {}'".format(cmd_prefix, macro)):
            return False
    return True

def getMockConfig(buildroot):
    rp_mock_img = getattr(buildroot.options, 'rp_mock_img', None)
    buildmode = getattr(buildroot.options, 'buildmode', SIMPLE)
    targetarches = getattr(buildroot.options, 'targetarches', None)
    optimization = getattr(buildroot.options, 'optimization', None)
    ccachedir = getattr(buildroot.options, 'ccachedir', None)
    sign = getattr(buildroot.options, 'sign', True)
    managed = getattr(buildroot.options, 'managed', True)
    nocleanup_build = getattr(buildroot.options, 'nocleanup_build', False)
    # Get the target arch cross or not
    native_arch = getArchCrossNative(buildroot.target_arch)
    target_arch = getArchCrossTarget(buildroot.target_arch)
    target_arch_rpm = get_arch_from_rpmmacro(target_arch)

    buildroot.logger.debug("####### IOTBZH: write config. Build Mode: %d, br_arch: %s" % (
        buildmode, buildroot.br_arch))

    opts = {}
    for k in ('repoid', 'tag_name'):
        if hasattr(buildroot, k):
            opts[k] = getattr(buildroot, k)
    for k in ('mockdir', 'topdir', 'topurl', 'topurls', 'packager', 'vendor',
              'distribution', 'mockhost', 'yum_proxy', 'rpmbuild_timeout'):
        if hasattr(buildroot.options, k):
            opts[k] = getattr(buildroot.options, k)
    opts['buildroot_id'] = buildroot.id
    if buildroot.setup_dns:
        opts['rpmbuild_networking'] = True
    opts['use_host_resolv'] = buildroot.setup_dns
    opts['install_group'] = buildroot.install_group
    opts['maven_opts'] = buildroot.maven_opts
    opts['maven_envs'] = buildroot.maven_envs
    opts['bind_opts'] = buildroot.bind_opts
    opts['target_arch'] = target_arch
    if 'mock.package_manager' in buildroot.config['extra']:
        opts['package_manager'] = buildroot.config['extra']['mock.package_manager']

    opts['plugin_conf'] = {}
    # force signing rpm if macros exists
    if sign and _check_signing(buildroot):
        opts['plugin_conf'] = {
            'sign_enable': True,
            'sign_opts': {
                'cmd': 'rpmsign',
                'opts': '--addsign %(rpms)s'
            }
        }

    # force bootstrap false
    opts['use_bootstrap'] = False
    # force module_hotfixes
    opts['module_hotfixes'] = 1
    if buildroot.internal_dev_setup is not None:
        opts['internal_dev_setup'] = bool(buildroot.internal_dev_setup)
    opts['tag_macros'] = {}
    for key in buildroot.config['extra']:
        if key.startswith('rpm.macro.'):
            macro = '%' + key[10:]
            opts['tag_macros'][macro] = buildroot.config['extra'][key]

    # ccache optimization from localbuilder
    if optimization:
        opts['plugin_conf']['nosync_force'] = True
        opts['plugin_conf']['ccache_enable'] = True
        opts['plugin_conf']['root_cache_enable'] = True
        opts['plugin_conf']['yum_cache_enable'] = True
        opts['plugin_conf']['package_state_enable'] = False
        opts['plugin_conf']['nosync'] = True

    # override repoid
    if getattr(buildroot.options, 'repolatest', False):
        opts['repoid'] = 'latest'
    #output = koji.genMockConfig(self.name, self.br_arch, managed=True, **opts)
    arch_repo = "$basearch"
    output = koji.genMockConfig(
        buildroot.name, arch_repo, managed=managed, **opts)

    # remove repo
    output = re.sub(r"(^config_opts\['yum.conf'\] = .*)# repos\\n\\n\[build\].*$", r"\1'", output, flags=re.MULTILINE)

    if nocleanup_build:
        output += "config_opts['exclude_from_homedir_cleanup'] = ['build/BUILD', 'build/SPECS', 'build/SOURCES', 'build/RPMS']" + os.linesep

    if ccachedir:
        output += os.linesep + \
            "config_opts['plugin_conf']['ccache_opts']['dir'] = '%s'" % ccachedir + os.linesep

    distprefix = getPrjDist(buildroot.tag_name)
    neg_tags = buildroot.session.reposInheritances(buildroot.tag_name, baseurl=buildroot.options.topurl, with_repoid=True)
    buildroot.logger.debug(
        '###### DistPrefix would be %s, neg_tags: %s' % (distprefix, neg_tags))
    output += os.linesep + "##### IotBzh append macros #####" + os.linesep
    if distprefix:
        output += "config_opts['macros']['%distprefix'] = '" + \
            distprefix + "'" + os.linesep

    for neg_tag in neg_tags:
        neg_tag_name = neg_tag.get('name', re.sub(r"[$:/\.]", "", neg_tag['url']))
        output += os.linesep + "config_opts['yum.conf'] += '\\n[%s]\\nname=%s\\nbaseurl=%s\\npriority=%s\\nmodule_hotfixes=1'" % (
            neg_tag_name, neg_tag_name, neg_tag['url'], neg_tag['priority']) + os.linesep

    # forcearch if not armv7hl gives i686
    output += "config_opts['forcearch'] = '" + native_arch + "'" + os.linesep
    # force boot efi mode if requested
    if getattr(buildroot.options, 'boot_efi', None):
        output += "config_opts['files']['/tmp/blivet.storage.force.efi'] = 'True'" + os.linesep

    # accel mock config
    if buildmode == ACCEL and not isNative(target_arch):
        output += os.linesep + \
            '###### Accel mock config for %s' % target_arch + os.linesep
        output += "config_opts['chroot_additional_packages'] = " + \
                  "'" + "qemu-accel-" + target_arch_rpm + "'" + os.linesep

    if buildmode in (CROSS_SIMPLE, CROSS_COMPILERS):
        output += os.linesep + \
            '###### Write common cross mock config for %s' % target_arch + os.linesep

        # Cross compilers are quite special as we want to get a Cross BuildRoot
        # to build normally. So we need to get the native arch
        if buildmode == CROSS_COMPILERS:
            buildroot.logger.debug('###### Write compiler cross mock config')
            # config_opts['rpmbuild_arch'] is set at
            # config_opts['target_arch'] by default
            output += "config_opts['targetarches'] = {}".format(
                targetarches) + os.linesep
        else:
            output += "config_opts['plugin_conf']['check_binarch_enable'] = True" + os.linesep
            output += "config_opts['rpmbuild_arch'] = '" + \
                target_arch + "'" + os.linesep
            pkgs_cross = ""
            pkgs_cross += " cross-profile-setup-" + target_arch
            if getattr(buildroot.options, 'pkg_name',
                       'not-ghc').startswith('ghc-'):
                pkgs_cross += " cross-ghc-compiler-" + target_arch_rpm

            pkgs_cross += " gcc-" + target_arch_rpm + "-linux-gnu"
            pkgs_cross += " gcc-c++-" + target_arch_rpm + "-linux-gnu"
            # pkgs_cross += " gcc-gccgo-" + target_arch_rpm + "-linux-gnu"
            # pkgs_cross += " gcc-gfortran-" + target_arch_rpm + "-linux-gnu"
            # pkgs_cross += " gcc-gnat-" + target_arch_rpm + "-linux-gnu"

            # pkgs_cross += " qemu-user-static"

            output += "config_opts['chroot_additional_packages'] = " + \
                "'" + pkgs_cross + "'" + os.linesep

        output += "config_opts['macros']['%_crosstarget'] = '" + \
            target_arch_rpm + "-linux-gnu'" + os.linesep
        output += "config_opts['macros']['%_crossroot'] = '/usr/" + \
            target_arch_rpm + "-linux-gnu/sys-root'" + os.linesep
        output += "config_opts['environment']['QEMU_LD_PREFIX'] = '/usr/" + \
            target_arch_rpm + "-linux-gnu/sys-root/'" + os.linesep
        output += "config_opts['environment']['DBUS_SERVICES_DIR'] = '/usr/share/dbus-1/services'" + os.linesep
        output += "config_opts['plugin_conf']['crossbuild_enable'] = True" + os.linesep
        output += "config_opts['plugin_conf']['bind_mount_enable'] = True " + os.linesep
        output += "config_opts['plugin_conf']['bind_mount_opts']['dirs'].append(('/proc/sys/fs/binfmt_misc', os.path.join(config_opts['basedir'], config_opts['root'], '/root/proc/sys/fs/binfmt_misc') ))" + os.linesep

    if rp_mock_img:
        output += "config_opts['redhat_subscription_required'] = False" + os.linesep
        output += "config_opts['internal_dev_setup'] = True" + os.linesep
        output += "config_opts['plugin_conf']['bind_mount_enable'] = True" + os.linesep
        output += "config_opts['plugin_conf']['bind_mount_opts']['dirs'].append(('/dev', '/dev/' ))" + os.linesep
        output += "config_opts['use_nspawn'] = False" + os.linesep

    return output


class GenMockConfig:
    def __init__(self, session, options,
                 kojid_path="/usr/sbin/kojid",
                 crossbuild_path="/usr/lib/koji-builder-plugins/crossbuild.py"):
        self._genmockconfig = None
        self._session = session
        self._session.host = Host()
        self._options = options

        # hotfix
        self._options.mock_bootstrap_image = False

        try:
            # ignore some unused import in our case
            self._fake_load_modules(('Cheetah', 'Cheetah.Template', 'multilib',
                                     'multilib.multilib', 'pykickstart',
                                     'pykickstart.handlers',
                                     'pykickstart.handlers.control'))

            self._kojid = load_source('kojid', kojid_path)
            self._kojid.BuildRoot._writeMockConfig = writeMockConfigIgnore

            sys.modules["__main__"].ImageTask = self._kojid.ImageTask
            sys.modules["__main__"].OzImageTask = self._kojid.OzImageTask
            sys.modules["__main__"].BuildRoot = self._kojid.BuildRoot
            sys.modules["__main__"].BuildTask = self._kojid.BuildTask
            sys.modules["__main__"].BuildImageTask = self._kojid.BuildImageTask
            sys.modules["__main__"].BuildBaseImageTask = self._kojid.BuildBaseImageTask
            sys.modules["__main__"].BuildArchTask = self._kojid.BuildArchTask
            sys.modules["__main__"].BuildSRPMFromSCMTask = self._kojid.BuildSRPMFromSCMTask
            sys.modules["__main__"].CreaterepoTask = self._kojid.CreaterepoTask

            self._crossbuild = load_source("crossbuild", crossbuild_path)
        except:
            raise
            self._genmockconfig = False
        else:
            self._genmockconfig = True

    def _fake_load_modules(self, modules):
        # need to load module + submodule (as attribute as parent module)
        for name in modules:
            submodules = name.split(".")

            parent_module = None
            while submodules:
                module_name = submodules.pop(0)
                full_module_name = parent_module.__name__ + "." + \
                    module_name if parent_module else module_name
                module = sys.modules.get(full_module_name, None)

                if not module:
                    module = types.ModuleType(full_module_name)
                    # set submodule as parent attribute
                    if parent_module:
                        setattr(parent_module, module_name, module)
                    # load module into sys
                    sys.modules[full_module_name] = module

                parent_module = module

    def buildrootConfig(self, tag, arch, *arg, **kwargs):
        if not self._genmockconfig:
            return None

        repo = self._session.getRepo(tag)
        # no repo: because the project was just created
        #          and the newRepo is not done
        #          so try to do a newRepo and wait it is done
        # TODO: should not be called when the first newRepo is not done
        if not repo:
            taskid = self._session.newRepo(tag)
            state = koji.TASK_STATES['FREE']
            while state in (koji.TASK_STATES['FREE'], koji.TASK_STATES['OPEN']):
                time.sleep(1)
                state = self._session.getTaskInfo(taskid).get("state", None)
            repo = self._session.getRepo(tag)
            if not repo:
                raise MockConfigRedError(
                        f"No repo found for tag:{tag} arch:{arch}: "
                        "could be due to a newRepo not finished when project is created")

        # default mode is cross for non native architectures
        if getattr(self._options, 'buildmode', -1) < 0 and not isNative(arch):
            self._options.buildmode = CROSS_SIMPLE

        # getMockConfig needs cross arches
        if self._options.buildmode is CROSS_SIMPLE and not isCross(arch):
            arch = getCrossArch(arch)

        buildroot = self._kojid.BuildRoot(
            self._session, self._options, tag, arch, 1, repo["id"], *arg, **kwargs)

        # name of the 'root': mock buildroot
        mockrootname = getattr(self._options, 'mockrootname', None)
        buildroot.name = f"{tag}-{arch}" if not mockrootname else mockrootname
        #TODO: UGLYFIX there is an issue in iotpkg, that replace x86_64 by arch
        # for now remove the first character of the arch
        buildroot.name = buildroot.name.replace("x86_64", "x8664")

        return getMockConfig(buildroot)

class FakeOption:
    def __init__(self) -> None:
        self.mock_bootstrap_image = None
        self.buildmode = 0
        self.managed = False
        self.sign = False
        self.mockrootname = None
        self.packager = "redpesk"
        self.vendor = "redpesk"
        self.distribution = "unkown"
        self.topurl = None
        self.optimization = True
        self.ccache_dir = '%(cache_topdir)s/ccache/'
        self.nocleanup_build = True

class FakeSession:
    ERROR_MSG = "should be a list of repos matching '[{\"priority\": int, \"url\": str}]'"
    def __init__(self, repofile, distro) -> None:
        reposdistros = None
        with open(repofile, 'r') as f:
            reposdistros = json.loads(f.read())

        if not reposdistros:
            raise self._err_json_format(repofile)

        if not isinstance(reposdistros, dict):
            raise self._err_json_format(repofile)

        self.repos = reposdistros.get(distro, None)
        if not self.repos:
            raise self._err_json_format(repofile)

        if not isinstance(self.repos, list):
            raise self._err_json_format(repofile)

        repo_url = self.repos[0].get('url', None)
        repo_priority = self.repos[0].get('priority', None)
        if not (repo_url and repo_priority != None):
            raise self._err_json_format(repofile)

    def _err_json_format(self, repofile):
        raise Exception(f"repofile {repofile} is not correct {FakeSession.ERROR_MSG}")

    def getBuildConfig(self, tag, event=None):
        return {
            "config": "fakeconfig",
            "id": "faketagid",
            "name": "fakeconfigname",
            "extra": {}
        }

    def getRepo(self, tag):
        return {
            "id": "fakerepoid",
            }

    def repoInfo(self, repo_id, strict=True):
        return {
            "id": "fakerepoid",
            "create_event": "fakeevent",
            "tag_id": "faketagid",
            "tag_name": "faketagname",
            "state": 1,
            }

    def reposInheritances(self, *args, **kwargs):
        return self.repos
