
import copy
import json
import typing

import redpesk_utils.log
import redpesk_utils.type
import redpesk_utils.error
import redpesk_utils.common


class Parameter:
    def __init__(
            self,
            handle: str,
            value: typing.Any,
            type: redpesk_utils.type.Type,
            name: str = str(),
            description: str = str(),
            values: typing.List[typing.Any] = list()) -> None:
        self._type = type
        self._value = value
        self._values = values
        self._handle = handle
        if not name:
            name = handle.capitalize()
        self._name = name
        if not description:
            description = name
        self._description = description

    @property
    def handle(self) -> str:
        return self._handle

    @property
    def type(self) -> redpesk_utils.type.Type:
        return self._type

    @property
    def description(self) -> str:
        return self._description
    
    @property
    def values(self) -> typing.List[typing.Any]:
        return self._values

    @property
    def value(self) -> typing.Any:
        return self._value


    @value.setter
    def value(self, value: typing.Any) -> None:
        self._value = value

    def check_value_type(self) -> typing.Any:
        ok = False
        if self._type == redpesk_utils.type.Type.int:
            ok = isinstance(self._value, int)
        elif self._type == redpesk_utils.type.Type.float:
            ok = isinstance(self._value, float)
        elif self._type == redpesk_utils.type.Type.bool:
            ok = isinstance(self._value, bool)
        elif self._type == redpesk_utils.type.Type.string:
            ok = isinstance(self._value, str)
            if ok:
                self._value = str.strip(self._value)
        elif self._type == redpesk_utils.type.Type.string_list:
            ok = isinstance(self._value, str) and self._value in self._values
            if ok:
                self._value = str.strip(self._value)
        elif self._type == redpesk_utils.type.Type.string_array:
            ok = (isinstance(self._value, list) and
                  all(isinstance(elem, str) for elem in self._value))
            if ok:
                self._value = list(map(str.strip, self._value))
        if not ok:
            raise redpesk_utils.error.ParameterError(redpesk_utils.log.format(
                self._name, f"got {self._value} ({type(self._value)}), expected {self._type}" + (f" ({self._values})" if self._values else str())))
        return self._value

    def get_info(self) -> typing.Tuple[str, dict]:
        info = {redpesk_utils.common.FIELD_NAME: self._name,
                redpesk_utils.common.FIELD_VALUE: self._value,
                redpesk_utils.common.FIELD_VALUES: self._values,
                redpesk_utils.common.FIELD_TYPE: self._type.value,
                redpesk_utils.common.FIELD_DESCRIPTION: self._description}
        return self._handle, info


class Metadata:
    def __init__(self) -> None:
        self._parameters: typing.List[Parameter] = list()
        self._parameters.extend([
            Parameter(
                value=str(),
                type=redpesk_utils.type.Type.string,
                handle=redpesk_utils.common.FIELD_REVISION,
                description=redpesk_utils.common.METADATA_REVISION_DESCRIPTION),
            Parameter(
                value=str(),
                type=redpesk_utils.type.Type.string,
                handle=redpesk_utils.common.FIELD_SOURCES_URI,
                description=redpesk_utils.common.METADATA_SOURCE_URI_DESCRIPTION),
            Parameter(
                type=redpesk_utils.type.Type.string_list,
                handle=redpesk_utils.common.TYPE_BUILD_TYPE,
                value=redpesk_utils.common.BUILD_TYPE_DEFAULT,
                description=redpesk_utils.common.METADATA_BUILD_TYPE_DESCRIPTION,
                values=list(redpesk_utils.common.BUILD_TYPE_TRANSLATION.keys())+list(redpesk_utils.common.BUILD_TYPE_TRANSLATION.values())),
            Parameter(
                value=str(),
                type=redpesk_utils.type.Type.string,
                handle=redpesk_utils.common.FIELD_SPECFILE_LOCATION,
                description=redpesk_utils.common.METADATA_SPECFILE_LOCATION_DESCRIPTION),
            Parameter(
                type=redpesk_utils.type.Type.string_list,
                handle=redpesk_utils.common.TYPE_SPECFILE_TYPE,
                value=redpesk_utils.common.SPECFILE_TYPE_DEFAULT,
                description=redpesk_utils.common.METADATA_SPECFILE_TYPE_DESCRIPTION,
                values=list(redpesk_utils.common.SPECFILE_TYPE_TRANSLATION.keys())+list(redpesk_utils.common.SPECFILE_TYPE_TRANSLATION.values()))
        ])

    @property
    def parameters(self) -> typing.List[Parameter]:
        return self._parameters

    def get_key(self, key: str) -> Parameter:
        return next(
            parameter for parameter in self._parameters if key == parameter.handle)

    def format(self, content: str, strict: bool) -> str:
        content: typing.Dict[str, typing.Any] = json.loads(content)
        self.process(content, strict)
        return redpesk_utils.common.encode_dict(content)

    def process(self, content: typing.Dict[str, typing.Any], strict: bool) -> None:
        if content[redpesk_utils.common.FIELD_VERSION] != METADATA_TEMPLATE[redpesk_utils.common.FIELD_VERSION]:
            content[redpesk_utils.common.FIELD_VERSION] = METADATA_TEMPLATE[redpesk_utils.common.FIELD_VERSION]
            for data in list_key(copy.deepcopy(content)):
                if data not in list_key(METADATA_TEMPLATE):
                    pop_key(content, data)
            for data in list_key(METADATA_TEMPLATE):
                if data not in list_key(content):
                    set_key(content, data, get_key(METADATA_TEMPLATE, data))
        for key in list_key(content):
            parameter = self.get_key(key)
            parameter.value = get_key(content, key)
            try:
                parameter.check_value_type()
            except redpesk_utils.error.ParameterError as error:
                if not strict:
                    parameter.value = get_key(METADATA_TEMPLATE, key)
                    redpesk_utils.log.red_log.warning(
                        redpesk_utils.log.format(
                            error, f"{key} = {parameter.value}"))
                else:
                    raise error
            set_key(content, key, parameter.value)


def list_key(content: typing.Dict[str, typing.Any]) -> typing.List[str]:
    return content[redpesk_utils.common.FIELD_DATA].keys()


def get_key(content: typing.Dict[str, typing.Any], key: str) -> typing.Any:
    return content[redpesk_utils.common.FIELD_DATA][key]


def set_key(content: typing.Dict[str, typing.Any],
            key: str, value: typing.Any) -> None:
    if key == redpesk_utils.common.TYPE_BUILD_TYPE:
        value = redpesk_utils.common.BUILD_TYPE_LONG(value)
    elif key == redpesk_utils.common.TYPE_SPECFILE_TYPE:
        value = redpesk_utils.common.SPECFILE_TYPE_LONG(value)
    content[redpesk_utils.common.FIELD_DATA][key] = value


def pop_key(content: typing.Dict[str, typing.Any], key: str) -> None:
    content[redpesk_utils.common.FIELD_DATA].pop(key)


def autoversion(content: str) -> bool:
    return json.loads(content).get(
        "setverrel",
        dict()).get(
        redpesk_utils.common.FIELD_PARAMETERS,
        dict()).get(
            "version",
            dict()).get(
                redpesk_utils.common.FIELD_VALUE,
        False) if content else False


def autorelease(content: str) -> bool:
    return json.loads(content).get(
        "setverrel",
        dict()).get(
        redpesk_utils.common.FIELD_PARAMETERS,
        dict()).get(
            "release",
            dict()).get(
                redpesk_utils.common.FIELD_VALUE,
        False) if content else False


def template() -> typing.Dict[str, typing.Any]:
    template: typing.Dict[str, typing.Any] = dict()
    template[redpesk_utils.common.FIELD_DATA] = dict()
    template[redpesk_utils.common.FIELD_VERSION] = redpesk_utils.common.METADATA_VERSION
    [
        set_key(
            template,
            parameter.handle,
            parameter.value) for parameter in METADATA.parameters]
    redpesk_utils.log.red_log.debug(
        redpesk_utils.log.format(
            redpesk_utils.common.LOGGING_METADATA,
            redpesk_utils.common.METADATA_VERSION))
    return template


METADATA = Metadata()
METADATA_TEMPLATE = {}
