import io
import os
import re
import rpm
import typing
import logging
import tempfile
import contextlib

import redpesk_utils
import redpesk_utils.log
import redpesk_utils.error
import redpesk_utils.common
import redpesk_utils.gitpkg
import redpesk_utils.parameter
import redpesk_utils.redirectoutputs
import redpesk_utils.crossutils

SOURCEDIR = os.path.expanduser("~/rpmbuild/SOURCES/")


def write_autogen(
        gitpkg: redpesk_utils.gitpkg.Gitpkg,
        content: str,
        autogen: str,
        service: str) -> str:
    specfile: redpesk_utils.Spec = None
    if autogen:
        specfile: redpesk_utils.Spec = redpesk_utils.from_string(autogen)
    else:
        specfile = redpesk_utils.from_string(content)
    orig = redpesk_utils.from_string(content)
    autogen = re.sub(
        rf"((?:^|\n){redpesk_utils.common.MACRO_NAME}:\s*)\S+?(?=\n|$)",
        f"\\g<1>{gitpkg.gitpkg}",
        content, 1, re.IGNORECASE)
    autogen = re.sub(
        rf"(?:^|\n){redpesk_utils.common.MACRO_HEXSHA}:\s*\S+?(?=\n|$)",
        str(), autogen, flags=re.IGNORECASE)
    autogen = re.sub(
        rf"(^|\n)({redpesk_utils.common.MACRO_VERSION})(:\s*)\S+?(?=\n|$)",
        f"\\g<1>{redpesk_utils.common.MACRO_HEXSHA}\\g<3>{specfile.sha}\n\\g<2>\\g<3>{specfile.version}",
        autogen, 1, re.IGNORECASE)
    if not redpesk_utils.parameter.autoversion(service):
        autogen = re.sub(
            rf"((?:^|\n){redpesk_utils.common.MACRO_VERSION}:\s*)\S+?(?=\n|$)",
            f"\\g<1>{orig.version}",
            autogen, 1, re.IGNORECASE)
    return autogen


def parse(content: str, retry: int = 60, create_sources: bool = False) -> rpm.spec:
    # iotbzh macros fix
    for macro, value in redpesk_utils.crossutils.RPMS_CROSS_MACROS.items():
        rpm.addMacro(macro, value)

    # ignore changelog issues when loading specfile
    rpm.addMacro('changelog', '#changelog')

    with tempfile.NamedTemporaryFile() as tmp:
        with open(tmp.name, 'w') as spec:
            spec.write(content)
        with tempSources() as sources:
            while True:
                with io.BytesIO() as stderr:
                    try:
                        with redpesk_utils.redirectoutputs.redirect_stderr(stderr):
                            return rpm.spec(tmp.name)
                    except Exception as rpmerr:
                        stderr = stderr.getvalue().decode('utf-8')
                        if not retry:
                            # Prevent parse fail for ineffective addMacro
                            return
                        matchTag: re.Match = re.search(r"Unknown tag:\s*%?(\S*)", stderr, re.IGNORECASE)
                        matchSource: re.Match = re.search(rf'.*({SOURCEDIR}\S*):.*', stderr, re.IGNORECASE)
                        if matchTag:
                            rpm.addMacro(matchTag.group(1), f"#{matchTag.group(1)}")
                            redpesk_utils.log.red_log.info(f"{str(retry)} - {rpm.addMacro.__name__}: {matchTag.group(1)}")
                        elif create_sources and matchSource:
                            open(matchSource.group(1), 'a').close()
                            sources.append(matchSource.group(1))
                            redpesk_utils.log.red_log.info(f"{str(retry)} - {matchSource.group(1)}")
                        else:
                            raise redpesk_utils.error.SpecfileError(f"{str(rpmerr).strip()} {stderr.strip()}")
                        retry -= 1

@contextlib.contextmanager
def tempSources() -> None:
    os.makedirs(SOURCEDIR, exist_ok=True)
    sources: typing.List = list()
    try:
        yield sources
    finally:
        for source in sources:
            os.remove(source)
